/* (c) 1996-2001, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* Hash File Authentication Database Interface
 *
 * This module supports an authentication database formatted as required by
 * Apache mod_auth_db and mod_auth_dbm.  The AUTH_DBM file uses the user's
 * login name (not including the terminating null) as a key, and contains
 * the encrypted user passwords (including a terminating null).
 *
 * If SHARE_AUTH_IDENT is defined, the same file also contains ident
 * information. The key is still the same, but the data is a newline and
 * null terminated string like:
 *   <passwd>:<uid>:<gid>:<fullname>:<directory>:<email>:<valcode>:\n
 */


#include "common.h"
#include "authident.h"
#include "hashfile.h"

#include <ctype.h>

#include "set_uid.h"
#include "entropy.h"
#include "udb.h"
#include "str.h"
#include "lock.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif

char *showopt_auth_module= "auth_db/auth_dbm (db="AUTH_DBM")";


#ifdef SHARE_AUTH_IDENT
int mknamdbent(char *bf, char *passwd, uid_t uid, gid_t gid,
	       char *gecos, char *dir, char *email, char *valcode,
	       int priv, int epriv);
#ifdef DBM_UID_FILE
int mkuiddbent(char *bf, char *name, char *passwd, gid_t gid,
	       char *gecos, char *dir, char *email, char *valcode,
	       int priv, int epriv);
#endif
#endif


/* AUTH_GETPASS - Get the encrypted password for a user.  Returns NULL if there
 * is no such user.
 */

#ifdef HAVEAUTH_GETPASS
char *auth_getpass(char *nam)
{
    static char pass[40];
    int rc;

    if (nam == NULL) return NULL;

    set_httpd();
    rc= fetchdbm(AUTH_DBM, nam, strlen(nam), pass, 40);
    set_cfadm();

    switch (rc)
    {
    case 0:  return pass;
    /*case 1:  die("auth_getpass cannot open db "AUTH_DBM" to read");*/
    default: return NULL;
    }
}
#endif


/* AUTH_WALK() - Get the first or next name out of the dbm file.
 */

#ifdef HAVEAUTH_WALK
char *auth_walk(int next)
{
    static char bf[MAX_LOGIN_LEN+1];

    if (next)
    {
	if (nextdbm(bf, MAX_LOGIN_LEN+1))
	    return NULL;
	else
	    return bf;
    }
    else
    {
	set_httpd();
	if (walkdbm(AUTH_DBM))
	{
	    set_cfadm();
	    die("auth_walk cannot open db "AUTH_DBM" to read");
	}
	set_cfadm();

	if (firstdbm(bf, MAX_LOGIN_LEN+1))
	    return NULL;
	else
	    return bf;
    }
}
#endif


/* AUTH_SEEK() - Seek to the named position in the dbm file.
 */

#ifdef HAVEAUTH_SEEK
void auth_seek(char *login)
{
    set_httpd();
    if (walkdbm(AUTH_DBM))
    {
	set_cfadm();
	die("auth_seek cannot open db "AUTH_DBM" to read");
    }
    set_cfadm();

    seekdbm(login, strlen(login));
}
#endif


/* AUTH_NEWPASS - change the user's password to the given value.  Pass in
 * the encrypted password.
 */

#ifdef HAVEAUTH_NEWPASS
void auth_newpass(struct btpasswd *pw, char *encpw)
{
#if defined(SHARE_AUTH_IDENT) || defined(DBM_UID_FILE)
    char bf[BFSZ];
    int len;
#endif
    int rc;

    set_httpd();

#ifdef SHARE_AUTH_IDENT
    len= mknamdbent(bf, encpw, pw->uid, pw->gid, pw->gecos, pw->dir,
	            pw->email, pw->valcode);

    rc= storedbm(AUTH_DBM,
    		 pw->name, strlen(pw->name),
    		 bf, len+1,
    		 0);
#else
    rc= storedbm(AUTH_DBM,
    		 pw->name, strlen(pw->name),
    		 encpw, strlen(encpw)+1,
    		 0);
#endif /* SHARE_AUTH_IDENT */
    if (rc)
    {
	set_cfadm();
	die("auth_newpass cannot open user database "AUTH_DBM);
    }

#ifdef DBM_UID_FILE
    len= mkuiddbent(bf, pw->passwd, pw->name, pw->gid, newfname, pw->dir,
	    pw->email, pw->valcode);

    if (storedbm(DBM_UID_FILE,
		 (char *)&pw->uid, sizeof(uid_t),
		 bf, len+1, 0))
    {
	set_cfadm();
	die("auth_newpass cannot open user database "DBM_UID_FILE);
    }
#endif

    set_cfadm();
}
#endif /* HAVEAUTH_NEWPASS */


/* AUTH_ADD - Add an account to the authentication database. */

#ifdef HAVEAUTH_ADD
#ifdef SHARE_AUTH_IDENT

void auth_add(char *lid, char *pw, int status,
    uid_t uid, gid_t gid, char *fn, char *dir, char *email, int valid,
    int priv, int epriv)
{
    char bf[BFSZ];
    char *valcode= NULL;
    char *encpw= mkpasswd(pw,status);
    int len;

    if (email != NULL && email[0] != '\0' && !valid)
	valcode= get_noise();

    len= mknamdbent(bf, encpw, uid, gid, fn, dir, email, valcode, priv, epriv);

    set_httpd();
    if (storedbm(AUTH_DBM,
                 lid, strlen(lid),      /* don't include terminating NUL */
                 bf, len+1,		/* include terminating NUL */
                 1))			/* create file if necessary */
    {
        set_cfadm();
        die("auth_add cannot open user database "AUTH_DBM" (%s)",errdbm());
    }

#ifdef DBM_UID_FILE
    len= mkuiddbent(bf, encpw, lid, gid, fn, dir, email, valcode, priv, epriv);

    if (storedbm(DBM_UID_FILE,
                 (char *)&uid, sizeof(uid_t),
                 bf, len+1,             /* include terminating NUL */
                 1))                    /* create file if necessary */
    {
        set_cfadm();
        die("Cannot open user database "DBM_UID_FILE" (%s)",errdbm());
    }
#endif
    set_cfadm();
}

#else

void auth_add(char *lid, char *pw, int status,
    uid_t uid, gid_t gid, char *fn, char *dir, char *email, int valid,
    int priv, int epriv)
{
    char *data= mkpasswd(pw,status);

    set_httpd();
    if (storedbm(AUTH_DBM,
                 lid, strlen(lid),      /* Don't include terminating NUL */
                 data, strlen(data)+1,  /* Include terminating NUL */
                 1))                    /* Create file if necessary */
    {
        set_cfadm();
	die("auth_add cannot open user database "AUTH_DBM" (%s)",errdbm());
    }
    set_cfadm();
}

#endif /* !SHARE_AUTH_IDENT */
#endif /* HAVEAUTH_ADD */


/* AUTH_DEL - Delete the named user
 */

#ifdef HAVEAUTH_DEL
int auth_del(char *login)
{
    int rc;

    set_httpd();
    rc= deletedbm(AUTH_DBM, login, strlen(login), 0);
    set_cfadm();

    return rc;
}
#endif
