/* (c) 1996-2001, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* Yapp-Compatible Identification Database Interface
 *
 * This supports an identity database compatible with Yapp installations that
 * do *not* use Unix logins.  Yapp maintains a file in the user's home
 * directory called 'passwd'.  This file contains one line:
 *
 *     <login>:<full name>:<email-address>
 *
 * With this module installed, Backtalk expects all users to have such a file,
 * and uses it as the place to store full names and email addresses.
 *
 * It will also create a shortened version of the standard Backtalk 'id'
 * tag file, containing only the user's uid number, gid number, and validation
 * code.  If a user is encountered who does not have an 'id' file, we create
 * one, assigning the USER group id and a new uid.
 *
 * As in ident_tagfile, the directory name is generated from the login name.
 */


#include "backtalk.h"
#include "authident.h"
#include "nextuid.h"
#include "group.h"
#include "entropy.h"
#include "set_uid.h"

#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif

#include "set_uid.h"
#include "str.h"
#include "udb.h"
#include "lock.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif

#define YAPP_PASSWD_FILE "passwd"

char *showopt_ident_module= "ident_yapp";

/* SAVE_TAG - Write the given variable/value pair into a tag file, replacing
 * the old value, if there was one, or appending the value.  The value is
 * not changed if it was protected and the user is not an administrator, in
 * which case a non-zero value is returned.   This always takes four arguments,
 * but the last may be either an integer or a string depending on the type.
 */

#if __STDC__
int save_tag(char *fname, int amadm, char *varname, int type, ...)
#else
int save_tag(fname, amadm, varname, type, va_alist)
    char *fname;
    int amadm;
    char *varname;
    int type;
    va_dcl
#endif
{
    va_list ap;
    char *bf, *val, *end, *var;
    FILE *fp;
    int i, protect, done= 0, rc= 0;

    bf= read_file(fname);
    VA_START(ap, type);

    /* Truncate and prepare to rewrite */
    set_httpd_if_httpd_home();
    if ((fp= fopen(fname,"w+")) == NULL)
	die("Cannot open file %s to write",fname);
    lock_exclusive(fileno(fp),fname);

    /* Process each line of saved file contents, writing back or replacing */
    for (var= bf; var != NULL && *var != '\0'; var= end+1)
    {
	/* Parse file line */
	if (protect= (*var == '*')) var++;
	val= strchr(var,'=');
	end= strchr(var,'\n');
	if (val != NULL && end != NULL && val > var && end > val)
	{
	    *(val++)= '\0';
	    *end= '\0';
	    if (!done && !strcmp(var,varname))
	    {
		done= 1;
		if (protect && !amadm)
		    rc= 1;
		else
		{
		    if (type == TK_INTEGER)
			fprintf(fp,"%s%s=%d\n", protect?"*":"",
			    var, va_arg(ap, int));
		    else
		    {
			if (protect) fputc('*',fp);
			wrtstrtag(fp,var,va_arg(ap, char *));
		    }
		    continue;
		}
	     }
	     fprintf(fp,"%s=%s\n", var-protect, val);
	}
    }

    /* Process the variable if not previously in file */
    if (!done)
    {
	if (type == TK_INTEGER)
	    fprintf(fp,"%s=%d\n", var, va_arg(ap, int));
	else
	    wrtstrtag(fp,var,va_arg(ap, char *));
    }

    fclose(fp);
    unlock(fname);
    set_cfadm_if_httpd_home();
    if (bf != NULL) free(bf);
    va_end(ap);
    return rc;
}


/* YF_CLEAN - replace colons and newlines in a string with spaces.  This is
 * a cleanup done before writing a name to a yapp passwd file.  Probably
 * mostly unnecessary, since the check routines do much the same.
 */

void yf_clean(char *p)
{
    if (p == NULL) return;
    while (*(p= firstin(p,":\n")) != '\0')
	*(p++)= ' ';
}


/* IDENT_GETPWNAM - Get the database entry for the named user.
 */

#ifdef HAVEIDENT_GETPWNAM
struct btpasswd *ident_getpwnam(char *nam)
{
    static struct btpasswd pw;
    static char stbf[BFSZ];
    char fname[BFSZ+1];
    char bf[BFSZ+1], *dir, *var, *val, *end, *b, *e, *stp;
    FILE *fp;
    int ch;

    pw.name= nam;
    pw.valcode= NULL;
    pw.priv= -1;
    pw.epriv= -1;
    stp= stbf;

    /* Construct the directory name */
    if (nam == NULL || (dir= homedirpath(nam)) == NULL)
	return NULL;
    pw.dir= dir;

    /* Open Yapp-ish passwd file */
    sprintf(fname,"%s/"YAPP_PASSWD_FILE,dir);
    set_httpd_if_httpd_home();
    fp= fopen(fname,"r");
    set_cfadm_if_httpd_home();
    if (fp == NULL) return NULL;

    b= fgets(bf,BFSZ,fp);
    fclose(fp);

    /* Various sanity checks */
    if (b == NULL || (e= strchr(b,':')) == NULL) return NULL;
    *e= '\0';
    if (strcmp(b,nam) || (e= strchr(b= e+1,':')) == NULL) return NULL;

    /* Get full name */
    *e= '\0';
    pw.gecos= stp;
    stp+= b - e + 1;
    if (stp > stbf + BFSZ) return NULL;
    strcpy(stp,b);

    /* Get email address */
    e= firstin(b= e+1,"\n\r ");
    *e= '\0'
    pw.email= stp;
    stp+= b - e + 1;
    if (stp > stbf + BFSZ) return NULL;
    strcpy(stp,b);

    /* Open tag file */
    sprintf(fname,"%s/"IDENT_FILE,dir);
    set_httpd_if_httpd_home();
    if ((fp= fopen(fname,"r")) == NULL)
    {
	/* Tag file is missing - make one assigning new uid */
	set_cfadm_if_httpd_home();
	pw.uid= next_uid(1);
	pw.gid= user_group_id();
	if (pw.email[0] == '\0')
	    pw.valcode= pw.email;
	else if (stp + 26 > stbf + BFSZ)
	    return NULL;
	else
	{
	    char *noise= get_noise();
	    pw.valcode= stp;
	    strcpy(pw.valcode,noise);
	    free(noise);
	}

	set_httpd_if_httpd_home();
	if ((fp= fopen(fname,"w")) == NULL)
	{
	    set_cfadm_if_httpd_home();
	    return NULL;
	}
	lock_exclusive(fileno(fp),fname);
	fprintf(fp,"*uid=%d\n*gid=%d\n", pw.uid, pw.gid);
	if (pw.valcode != NULL) wrtstrtag(fp, "valcode", pw.valcode);
	fclose(fp);
	unlock(fname);
	set_cfadm_if_httpd_home();
	return &pw;
    }
    lock_shared(fileno(fp),fname);
    
    /* Read through tag file */
    while (fgets(bf,BFSZ,fp) != NULL)
    {
	/* Parse apart the line */
	var= bf+ (bf[0] == '*');
	if ((val= strchr(var, '=')) == NULL)
	{
	    fclose(fp);
	    unlock(fname);
	    set_cfadm_if_httpd_home();
	    return NULL;
	}
	*(val++)= '\0';
	end= val + strlen(val) - 1;
	if (*end == '\n')
	    *(end--)= '\0';
	else
	    /* skip over rest of over-long line */
	    while ((ch= fgetc(fp)) != '\n' && ch != EOF)
		;

	if (!strcmp(var, "uid"))
	{
	    if (!isascii(*val) || !isdigit(*val)) goto error;
	    pw.uid= atoi(val);
	}
	else if (!strcmp(var, "gid"))
	{
	    if (!isascii(*val) || !isdigit(*val)) goto error;
	    pw.gid= atoi(val);
	}
	else if (!strcmp(var, "valcode"))
	{
	    if (*val != '(') goto error;
	    if (*end == ')') *end= '\0';
	    pw.valcode= stp;
	    stp= unesctag(val+1, pw.valcode, BFSZ - (stp - stbf), 0) + 1;
	    if (stp == NULL) goto error;
	}
	else if (!strcmp(var, "privacy"))
	{
	    if (!isascii(*val) || !isdigit(*val)) goto error;
	    pw.priv= atoi(val);
	}
	else if (!strcmp(var, "eprivacy"))
	{
	    if (!isascii(*val) || !isdigit(*val)) goto error;
	    pw.epriv= atoi(val);
	}
    }

    fclose(fp);
    unlock(fname);
    set_cfadm_if_httpd_home();
    return &pw;

error:
    fclose(fp);
    unlock(fname);
    set_cfadm_if_httpd_home();
    return NULL;
}
#endif


/* IDENT_NEWFNAME - change the user's fullname to the given value.
 */

#ifdef HAVEIDENT_NEWFNAME
void ident_newfname(struct btpasswd *pw, char *newfname)
{
    char *fname, *p;
    FILE *fp;

    /* Construct name of ident file */
    fname= gethomedir(pw->name);
    strcat(fname,"/");
    strcat(fname,YAPP_PASSWD_FILE);

    flushusercache();

    /* Clean up new full name */
    yf_clean(newfname);

    /* Write new passwd file */
    set_httpd_if_httpd_home();
    if ((fp= fopen(fname,"w")) == NULL)
	die("Cannot write to %s",fname);

    fprintf(fp,"%s:%s:%s\n",pw->name, newfname, pw->email);

    fclose(fp);
    set_cfadm_if_httpd_home();
}
#endif


/* IDENT_NEWEMAIL - change the user's email address to the given value.
 */

#ifdef HAVEIDENT_NEWEMAIL
void ident_newemail(struct btpasswd *pw, char *newemail, int valid)
{
    char *p, *dir;
    char fname[BFSZ];
    char *valcode= NULL;
    FILE *fp;

    /* Construct name of ident file */
    sprintf(fname,"%s/"YAPP_PASSWD_FILE,dir= gethomedir(pw->name));

    flushusercache();

    /* Clean up new email address */
    yf_clean(newemail);

    /* Write new passwd file */
    set_httpd_if_httpd_home();

    if ((fp= fopen(fname,"w")) == NULL)
	die("Cannot write to %s",fname);
    fprintf(fp,"%s:%s:%s\n",pw->name, pw->gecos, newemail);
    fclose(fp);

    if (newemail[0] != '\0' && !valid)
	valcode= get_noise();

    sprintf(fname,"%s/"IDENT_FILE,dir);
    save_tag(fname, 1, "valcode", TK_STRING, valcode == NULL ? "" : valcode);
    if (valcode != NULL) free(valcode);
}
#endif

/* IDENT_NEWGID - change the user's groupid to the given value.
 */

#ifdef HAVEIDENT_NEWGID
void ident_newgid(struct btpasswd *pw, gid_t newgid)
{
    char fname[BFSZ];

    /* Construct name of ident file */
    sprintf(fname,"%s/"IDENT_FILE,gethomedir(pw->name));

    flushusercache();

    save_tag(fname, 1, "gid", TK_INTEGER, newgid);
}
#endif

#ifdef HAVEIDENT_NEWPRIV
void ident_newpriv(struct btpasswd *pw, int priv, int epriv)
{
    char fname[BFSZ];

    /* Construct name of ident file */
    sprintf(fname,"%s/"IDENT_FILE,gethomedir(pw->name));

    flushusercache();

    /* A bit inefficient here - be nicer to do both in one pass */
    save_tag(fname, 1, "privacy", TK_INTEGER, priv);
    save_tag(fname, 1, "eprivacy", TK_INTERGER, epriv);
}
#endif


/* IDENT_ADD - Add an entry to the user database describing a user with
 * the given login, uid, gid, password, fullname, and directory.
 */

#ifdef HAVEIDENT_ADD
void ident_add(char *lid, uid_t uid, gid_t gid, char *fn, char *dir,
	char *email, int valid, int priv, int epriv)
{
    char fname[BFSZ];
    char *valcode= NULL;
    FILE *fp;

    yf_clean(fn);
    yf_clean(email);

    set_httpd_if_httpd_home();
    sprintf(fname,"%s/"YAPP_PASSWD_FILE,dir);
    if ((fp= fopen(fname,"w")) == NULL)
	die("could not create ident file %s",fname);
    fprintf(fp,"%s:%s:%s\n",lid,fn,(email == NULL) ? "" : email);
    fclose(fp);

    if (email != NULL && email[0] != '\0' && !valid)
	valcode= get_noise();

    sprintf(fname,"%s/"IDENT_FILE,dir);
    if ((fp= fopen(fname,"w")) == NULL)
	die("could not create ident file %s",fname);
    lock_exclusive(fileno(fp),fname);
    fprintf(fp,"*uid=%d\n*gid=%d\nvalcode=(%s)\n",
	    uid, gid, valcode == NULL ? "" : valcode);
    fprintf(fp,"*uid=%d\n*gid=%d\n", uid, gid);
    if (valcode != NULL) wrtstrtag(fp,"valcode",valcode);
    if (priv != -1) fprintf(fp,"privacy=%d\n", priv);
    if (epriv != -1) fprintf(fp,"eprivacy=%d\n", epriv);
    fclose(fp);
    unlock(fname);

    if (valcode != NULL) free(valcode);

    set_cfadm_if_httpd_home();
}
#endif


/* IDENT_DEL - Delete the named user from the ident database.  Actually, there
 * is nothing to do here, since we will always be deleting the directory
 * next, and that will take care of everything.
 */

#ifdef HAVEIDENT_DEL
void ident_del(char *name) { }
#endif
